////////////////////////////////////////////////////////////////////////////////
// chaos_image.cpp : image

////////////////////////////////////////////////////////////////////////////////
// include

#include "stdafx.h"

#include "chaos_image.h"

////////////////////////////////////////////////////////////////////////////////
// chaos_image : image class

////////////////////////////////////////////////////////////////////////////////
// create & destroy

// default constructor

chaos_image::chaos_image() :
	width(0), height(0),
	scanline_length(0),
	capacity(0),
	bits(NULL)
{
}

// destructor

chaos_image::~chaos_image() {
	if (bits != NULL) {
		free(bits);
		bits = NULL;
	}
}

// create image

bool chaos_image::create(
	int new_width, int new_height,
	bool clear_bits,
	chaos_uint32 clear_col)
{
	ASSERT((new_width > 0) && (new_height > 0));
	int new_scanline_length = new_width;
	int total_length = new_scanline_length*new_height;
	chaos_uint32* new_bits = bits;
	int bits_size = total_length*sizeof(chaos_uint32);
	if (bits == NULL) {
		new_bits = (chaos_uint32*)malloc(bits_size);
	} else if (total_length > capacity) {
		new_bits = (chaos_uint32*)realloc(bits, bits_size);
	}
	if (new_bits == NULL) {
		throw std::bad_alloc();
	}
	width = new_width;
	height = new_height;
	scanline_length = new_scanline_length;
	bits = new_bits;
	if (total_length > capacity) {
		capacity = total_length;
	}
	if (clear_bits) {
		fill(clear_col);
	}
	return true;
}

// destroy image

void chaos_image::destroy() {
	if (bits != NULL) {
		free(bits);
		bits = NULL;
	}
	width = 0;
	height = 0;
	scanline_length = 0;
	capacity = 0;
}

////////////////////////////////////////////////////////////////////////////////
// operation

// invert alpha

void chaos_image::invert_alpha() {
	chaos_uint8* inv_bits =
		(chaos_uint8*)get_bits(0, 0)+CHAOS_RGBA_ALPHA_OFS;
	int inv_step = (scanline_length-width)*4;
	int y = height;
	do {
		int x = width;
		do {
			*inv_bits = (chaos_uint8)(255-*inv_bits);
			inv_bits += 4;
		} while (--x != 0);
		inv_bits += inv_step;
	} while (--y != 0);
}

// fill

void chaos_image::fill(
	chaos_uint32 fill_col,
	const chaos_rect* clip_rect)
{
	chaos_rect fill_rect(0, 0, width, height);
	if (clip_rect != NULL) {
		fill_rect &= *clip_rect;
	}
	if (!fill_rect.empty()) {
		chaos_uint32* fill_bits =
			get_bits(fill_rect.get_lt_pos());
		int fill_step = scanline_length-fill_rect.width;
		if (fill_step == 0) {
			int red = CHAOS_RGBA_RED(fill_col),
				green = CHAOS_RGBA_GREEN(fill_col),
				blue = CHAOS_RGBA_BLUE(fill_col),
				alpha = CHAOS_RGBA_ALPHA(fill_col);
			if ((red == green) && (red == blue) && (red == alpha)) {
				memset(
					fill_bits, red,
					fill_rect.width*fill_rect.height*sizeof(chaos_uint32));
			} else {
				int i = fill_rect.width*fill_rect.height;
				do {
					*(fill_bits++) = fill_col;
				} while (--i != 0);
			}
		} else {
			int y = fill_rect.height;
			do {
				int x = fill_rect.width;
				do {
					*(fill_bits++) = fill_col;
				} while (--x != 0);
				fill_bits += fill_step;
			} while (--y != 0);
		}
	}
}

// fill width alpha

void chaos_image::alpha_fill(
	chaos_uint32 fill_col,
	const chaos_image* alpha_image,
	chaos_pos fill_pos,
	const chaos_rect* clip_rect)
{
	chaos_rect fill_rect(0, 0, width, height);
	fill_rect &= chaos_rect(fill_pos, alpha_image->get_dim());
	if (clip_rect != NULL) {
		fill_rect &= *clip_rect;
	}
	if (!fill_rect.empty()) {
		chaos_uint32* dst_bits =
			get_bits(fill_rect.get_lt_pos());
		const chaos_uint8* alpha_bits = (chaos_uint8*)
			alpha_image->get_bits(fill_rect.get_lt_pos()-fill_pos)+
				CHAOS_RGBA_ALPHA_OFS;
		int dst_step = scanline_length-fill_rect.width,
			alpha_step =
				(alpha_image->get_scanline_length()-fill_rect.width)*4;
		int fill_red = CHAOS_RGBA_RED(fill_col),
			fill_green = CHAOS_RGBA_GREEN(fill_col),
			fill_blue = CHAOS_RGBA_BLUE(fill_col);
		int y = fill_rect.height;
		do {
			int x = fill_rect.width;
			do {
				chaos_uint32 dst_col = *dst_bits;
				int alpha = *alpha_bits;
				alpha_bits += 4;
				alpha += (alpha+1)/256;
				int alpha2 = 256-alpha;
				*(dst_bits++) = CHAOS_RGBA(
					(CHAOS_RGBA_RED(dst_col)*alpha+
						fill_red*alpha2)/256,
					(CHAOS_RGBA_GREEN(dst_col)*alpha+
						fill_green*alpha2)/256,
					(CHAOS_RGBA_BLUE(dst_col)*alpha+
						fill_blue*alpha2)/256,
					0);
			} while (--x != 0);
			dst_bits += dst_step;
			alpha_bits += alpha_step;
		} while (--y != 0);
	}
}

// fade fill

void chaos_image::fade_fill(
	chaos_uint32 fill_col,
	int fade,
	const chaos_rect* clip_rect)
{
	if (fade >= 255) {
		return;
	}
	if (fade <= 0) {
		fill(fill_col, clip_rect);
		return;
	}
	int alpha = fade+(fade+1)/256;
	int alpha2 = 256-alpha;
	int fill_red = CHAOS_RGBA_RED(fill_col)*alpha2,
		fill_green = CHAOS_RGBA_GREEN(fill_col)*alpha2,
		fill_blue = CHAOS_RGBA_BLUE(fill_col)*alpha2;
	chaos_rect fill_rect(0, 0, width, height);
	if (clip_rect != NULL) {
		fill_rect &= *clip_rect;
	}
	if (!fill_rect.empty()) {
		chaos_uint32* dst_bits =
			get_bits(fill_rect.get_lt_pos());
		int dst_step = scanline_length-fill_rect.width;
		int y = fill_rect.height;
		do {
			int x = fill_rect.width;
			do {
				chaos_uint32 dst_col = *dst_bits;
				*(dst_bits++) = CHAOS_RGBA(
					(CHAOS_RGBA_RED(dst_col)*alpha+
						fill_red)/256,
					(CHAOS_RGBA_GREEN(dst_col)*alpha+
						fill_green)/256,
					(CHAOS_RGBA_BLUE(dst_col)*alpha+
						fill_blue)/256,
					0);
			} while (--x != 0);
			dst_bits += dst_step;
		} while (--y != 0);
	}
}

// transparent fill

void chaos_image::trans_fill(
	int trans_red, int trans_green, int trans_blue,
	const chaos_rect* clip_rect)
{
	chaos_rect fill_rect(0, 0, width, height);
	if (clip_rect != NULL) {
		fill_rect &= *clip_rect;
	}
	if (!fill_rect.empty()) {
		chaos_uint32* dst_bits =
			get_bits(fill_rect.get_lt_pos());
		int dst_step = scanline_length-fill_rect.width;
		int y = fill_rect.height;
		do {
			int x = fill_rect.width;
			do {
				chaos_uint32 dst_col = *dst_bits;
				int red = CHAOS_RGBA_RED(dst_col)+trans_red,
					green = CHAOS_RGBA_GREEN(dst_col)+trans_green,
					blue = CHAOS_RGBA_BLUE(dst_col)+trans_blue,
					alpha = CHAOS_RGBA_ALPHA(dst_col);
				if (red < 0) {
					red = 0;
				} else if (red > 255) {
					red = 255;
				}
				if (green < 0) {
					green = 0;
				} else if (green > 255) {
					green = 255;
				}
				if (blue < 0) {
					blue = 0;
				} else if (blue > 255) {
					blue = 255;
				}
				*(dst_bits++) = CHAOS_RGBA(red, green, blue, alpha);
			} while (--x != 0);
			dst_bits += dst_step;
		} while (--y != 0);
	}
}

// mono fill

void chaos_image::mono_fill(
	int mono_red, int mono_green, int mono_blue,
	const chaos_rect* clip_rect)
{
	chaos_rect fill_rect(0, 0, width, height);
	if (clip_rect != NULL) {
		fill_rect &= *clip_rect;
	}
	if (!fill_rect.empty()) {
		chaos_uint32* dst_bits =
			get_bits(fill_rect.get_lt_pos());
		int dst_step = scanline_length-fill_rect.width;
		int y = fill_rect.height;
		do {
			int x = fill_rect.width;
			do {
				chaos_uint32 dst_col = *dst_bits;
				int red = CHAOS_RGBA_RED(dst_col),
					green = CHAOS_RGBA_GREEN(dst_col),
					blue = CHAOS_RGBA_BLUE(dst_col),
					alpha = CHAOS_RGBA_ALPHA(dst_col);
				red = green = blue = (red+green+blue)/3;
				red += mono_red;
				green += mono_green;
				blue += mono_blue;
				if (red < 0) {
					red = 0;
				} else if (red > 255) {
					red = 255;
				}
				if (green < 0) {
					green = 0;
				} else if (green > 255) {
					green = 255;
				}
				if (blue < 0) {
					blue = 0;
				} else if (blue > 255) {
					blue = 255;
				}
				*(dst_bits++) = CHAOS_RGBA(red, green, blue, alpha);
			} while (--x != 0);
			dst_bits += dst_step;
		} while (--y != 0);
	}
}

// horizontal blind fill

void chaos_image::horz_blind_fill(
	chaos_uint32 fill_col,
	int gap, int fat,
	const chaos_rect* clip_rect)
{
	if ((gap == 0) || (fat <= 0)) {
		return;
	}
	int agap = abs(gap);
	if (fat >= agap) {
		fill(fill_col, clip_rect);
		return;
	}
	chaos_rect fill_rect(0, 0, width, height);
	if (clip_rect != NULL) {
		fill_rect &= *clip_rect;
	}
	if (!fill_rect.empty()) {
		chaos_uint32* dst_bits =
			get_bits(fill_rect.get_lt_pos());
		int dst_step = scanline_length-fill_rect.width;
		int y_start = fill_rect.y,
			y_end = fill_rect.y+fill_rect.height;
		if (gap > 0) {
			int y = y_start;
			do {
				if (y%agap < fat) {
					int x = fill_rect.width;
					do {
						*(dst_bits++) = fill_col;
					} while (--x > 0);
				} else {
					dst_bits += fill_rect.width;
				}
				dst_bits += dst_step;
			} while (++y < y_end);
		} else {
			int fat2 = agap-fat;
			int y = y_start;
			do {
				if (y%agap >= fat2) {
					int x = fill_rect.width;
					do {
						*(dst_bits++) = fill_col;
					} while (--x > 0);
				} else {
					dst_bits += fill_rect.width;
				}
				dst_bits += dst_step;
			} while (++y < y_end);
		}
	}
}

// vertical blind fill

void chaos_image::vert_blind_fill(
	chaos_uint32 fill_col,
	int gap, int fat,
	const chaos_rect* clip_rect)
{
	if ((gap == 0) || (fat <= 0)) {
		return;
	}
	int agap = abs(gap);
	if (fat >= agap) {
		fill(fill_col, clip_rect);
		return;
	}
	chaos_rect fill_rect(0, 0, width, height);
	if (clip_rect != NULL) {
		fill_rect &= *clip_rect;
	}
	if (!fill_rect.empty()) {
		chaos_uint32* dst_bits =
			get_bits(fill_rect.get_lt_pos());
		int dst_step = scanline_length-fill_rect.width;
		int x_start = fill_rect.x,
			x_end = fill_rect.x+fill_rect.width;
		if (gap > 0) {
			int y = fill_rect.height;
			do {
				int x = x_start;
				do {
					if (x%agap < fat) {
						*dst_bits = fill_col;
					}
					dst_bits++;
				} while (++x < x_end);
				dst_bits += dst_step;
			} while (--y > 0);
		} else {
			int fat2 = agap-fat;
			int y = fill_rect.height;
			do {
				int x = x_start;
				do {
					if (x%agap >= fat2) {
						*dst_bits = fill_col;
					}
					dst_bits++;
				} while (++x < x_end);
				dst_bits += dst_step;
			} while (--y > 0);
		}
	}
}

// blt

void chaos_image::blt(
	const chaos_image* blt_image,
	chaos_pos blt_pos,
	const chaos_rect* clip_rect)
{
	chaos_rect blt_rect(0, 0, width, height);
	blt_rect &= chaos_rect(blt_pos, blt_image->get_dim());
	if (clip_rect != NULL) {
		blt_rect &= *clip_rect;
	}
	if (!blt_rect.empty()) {
		chaos_uint32* dst_bits =
			get_bits(blt_rect.get_lt_pos());
		const chaos_uint32* src_bits =
			blt_image->get_bits(blt_rect.get_lt_pos()-blt_pos);
		int dst_step = scanline_length-blt_rect.width,
			src_step = blt_image->get_scanline_length()-blt_rect.width;
		if ((dst_step == 0) && (src_step == 0)) {
			memcpy(
				dst_bits, src_bits,
				blt_rect.width*blt_rect.height*sizeof(chaos_uint32));
		} else {
			int y = blt_rect.height;
			do {
				int x = blt_rect.width;
				do {
					*(dst_bits++) = *(src_bits++);
				} while (--x != 0);
				dst_bits += dst_step;
				src_bits += src_step;
			} while (--y != 0);
		}
	}
}

// stretch blt

void chaos_image::stretch_blt(
	const chaos_image* blt_image,
	chaos_pos blt_pos,
	chaos_dim blt_dim,
	const chaos_rect* clip_rect)
{
	chaos_dim src_dim = blt_image->get_dim();
	if (src_dim.empty()) {
		return;
	}
	if (src_dim == blt_dim) {
		blt(blt_image, blt_pos, clip_rect);
		return;
	}
	chaos_rect blt_rect(0, 0, width, height);
	blt_rect &= chaos_rect(blt_pos, blt_dim);
	if (clip_rect != NULL) {
		blt_rect &= *clip_rect;
	}
	if (!blt_rect.empty()) {
		chaos_uint32* dst_bits =
			get_bits(blt_rect.get_lt_pos());
		int dst_step = scanline_length-blt_rect.width;
		chaos_rect dst_rect = blt_rect-blt_pos;
		int dst_left = blt_rect.x-blt_pos.x,
			dst_top = blt_rect.y-blt_pos.y,
			dst_right = dst_left+dst_rect.width,
			dst_bottom = dst_top+dst_rect.height;
		for (int y = dst_top; y < dst_bottom; y++) {
			int y2 = (y*src_dim.height)/blt_dim.height;
			const chaos_uint32* src_bits =
				blt_image->get_bits(0, y2);
			for (int x = dst_left; x < dst_right; x++) {
				int x2 = (x*src_dim.width)/blt_dim.width;
				*(dst_bits++) = *(src_bits+x2);
			}
			dst_bits += dst_step;
		}
	}
}

// smooth stretch blt(zoom out only)

void chaos_image::smooth_stretch_blt(
	const chaos_image* blt_image,
	chaos_pos blt_pos,
	chaos_dim blt_dim,
	const chaos_rect* clip_rect)
{
	chaos_dim src_dim = blt_image->get_dim();
	if (src_dim.empty()) {
		return;
	}
	if ((src_dim.width <= blt_dim.width) ||
		(src_dim.height <= blt_dim.height))
	{
		stretch_blt(
			blt_image,
			blt_pos,
			blt_dim,
			clip_rect);
		return;
	}
	chaos_rect blt_rect(0, 0, width, height);
	blt_rect &= chaos_rect(blt_pos, blt_dim);
	if (clip_rect != NULL) {
		blt_rect &= *clip_rect;
	}
	if (!blt_rect.empty()) {
		chaos_uint32* dst_bits =
			get_bits(blt_rect.get_lt_pos());
		int dst_step = scanline_length-blt_rect.width;
		chaos_rect dst_rect = blt_rect-blt_pos;
		int dst_left = blt_rect.x-blt_pos.x,
			dst_top = blt_rect.y-blt_pos.y,
			dst_right = dst_left+dst_rect.width,
			dst_bottom = dst_top+dst_rect.height;
		int src_y1,
			src_y2 = (dst_top*src_dim.height)/blt_dim.height;
		for (int dst_y = dst_top; dst_y < dst_bottom; dst_y++) {
			src_y1 = src_y2;
			src_y2 = ((dst_y+1)*src_dim.height)/blt_dim.height;
			ASSERT((src_y1 >= 0) && (src_y1 < blt_image->height));
			ASSERT((src_y2 > src_y1) && (src_y2 <= blt_image->height));
			int src_x1,
				src_x2 = (dst_left*src_dim.width)/blt_dim.width;
			for (int dst_x = dst_left; dst_x < dst_right; dst_x++) {
				src_x1 = src_x2;
				src_x2 = ((dst_x+1)*src_dim.width)/blt_dim.width;
				ASSERT((src_x1 >= 0) && (src_x1 < blt_image->width));
				ASSERT((src_x2 > src_x1) && (src_x2 <= blt_image->width));
				int dot_count = (src_x2-src_x1)*(src_y2-src_y1);
				chaos_uint32 dst_col = CHAOS_RGBA(0, 0, 0, 0);
				if (dot_count > 0) {
					double red = 0, green = 0, blue = 0;
					const chaos_uint32* src_bits =
						blt_image->get_bits(src_x1, src_y1);
					int src_step = blt_image->get_scanline_length()-
						(src_x2-src_x1);
					for (int src_y = src_y1; src_y < src_y2; src_y++) {
						for (int src_x = src_x1; src_x < src_x2; src_x++) {
							chaos_uint32 src_col = *(src_bits++);
							red += CHAOS_RGBA_RED(src_col);
							green += CHAOS_RGBA_GREEN(src_col);
							blue += CHAOS_RGBA_BLUE(src_col);
						}
						src_bits += src_step;
					}
					dst_col = CHAOS_RGBA(
						(int)(red/dot_count+0.5),
						(int)(green/dot_count+0.5),
						(int)(blue/dot_count+0.5),
						0);
				}
				*(dst_bits++) = dst_col;
			}
			dst_bits += dst_step;
		}
	}
}

// setup for fast smooth stretch blt

void chaos_image::setup_fast_smooth_stretch_blt_info(
	fast_smooth_stretch_blt_info* bi,
	const chaos_rect& dst_rect,
	const chaos_rect& src_rect,
	const chaos_rect& clip_rect)
{
	static bool gamma_table_initialized = false;
	static chaos_uint32 gamma_table[GAMMA_TABLE_COUNT+1];
	if (!gamma_table_initialized) {
		gamma_table_initialized = true;
		for (int i = 0; i <= GAMMA_TABLE_COUNT; i++) {
			double f;
			f = (1-cos(i*3.14159265358979323846/GAMMA_TABLE_COUNT))/2;
			int n  = (int)(f*GAMMA_TABLE_VALUE+0.5);
			if (n < 0) {
				n = 0;
			} else if (n > GAMMA_TABLE_VALUE) {
				n = GAMMA_TABLE_VALUE;
			}
			gamma_table[i] = n;
		}
	}
	if ((dst_rect == bi->dst_rect) &&
		(src_rect == bi->src_rect) &&
		(clip_rect == bi->clip_rect))
	{
		return;
	}
	bi->clear();
	bi->dst_rect = dst_rect;
	bi->src_rect = src_rect;
	bi->clip_rect = clip_rect;
	bi->same_size = (dst_rect.get_dim() == src_rect.get_dim());
	bi->blt_src_rect = src_rect;
	if ((bi->blt_src_rect.width <= 0) ||
		(bi->blt_src_rect.height <= 0))
	{
		return;
	}
	bi->blt_dst_rect = dst_rect;
	bi->blt_dst_rect &= clip_rect;
	bi->blt_dst_rect &= chaos_rect(
		chaos_pos(
			((((bi->blt_src_rect.x-src_rect.x)*
					dst_rect.width)+src_rect.width-1)/
						src_rect.width)+dst_rect.x,
			((((bi->blt_src_rect.y-src_rect.y)*
					dst_rect.height)+src_rect.height-1)/
						src_rect.height)+dst_rect.y),
		chaos_pos(
			(((bi->blt_src_rect.get_right()-src_rect.x)*
				dst_rect.width)/
					src_rect.width)+dst_rect.x,
			(((bi->blt_src_rect.get_bottom()-src_rect.y)*
				dst_rect.height)/
					src_rect.height)+dst_rect.y));
	if ((bi->blt_dst_rect.width <= 0) ||
		(bi->blt_dst_rect.height <= 0))
	{
		return;
	}
	int draw_x = bi->blt_dst_rect.x-dst_rect.x,
		draw_y = bi->blt_dst_rect.y-dst_rect.y,
		draw_cx = bi->blt_dst_rect.width,
		draw_cy = bi->blt_dst_rect.height;
	bi->src_y_ofses.resize(draw_cy*2, 0);
	bi->gamma_y.resize(draw_cy, 0);
	if (dst_rect.height >= src_rect.height) {
		int base_y = src_rect.y*dst_rect.height*2,
			max_y = src_rect.height*dst_rect.height*2;
		for (int y = 0; y < draw_cy; y++) {
			int y1 = ((y+draw_y)*2+1)*src_rect.height+base_y,
				y2 = y1-dst_rect.height;
			if (y2 < 0) {
				y2 = 0;
			} else if (y2 >= max_y) {
				y2 = max_y-1;
			}
			int gamma = GAMMA_TABLE_COUNT-
				((y2%(dst_rect.height*2))*GAMMA_TABLE_COUNT)/
					(dst_rect.height*2);
			bi->gamma_y[y] = gamma_table[gamma];
			y2 /= dst_rect.height*2;
			bi->src_y_ofses[y*2] = y2;
			if (y2 < src_rect.height-1) {
				y2++;
			}
			bi->src_y_ofses[y*2+1] = y2;
		}
	} else {
		int base_y = src_rect.y*dst_rect.height*4,
			max_y = src_rect.height*dst_rect.height*4;
		for (int y = 0; y < draw_cy; y++) {
			int y1 = ((y+draw_y)*2+1)*src_rect.height*2+base_y,
				y2 = y1-src_rect.height,
				y3 = y1+src_rect.height;
			if (y2 < 0) {
				y2 = 0;
			} else if (y2 >= max_y) {
				y2 = max_y-1;
			}
			if (y3 < 0) {
				y3 = 0;
			} else if (y3 >= max_y) {
				y3 = max_y-1;
			}
			y2 /= dst_rect.height*4;
			bi->src_y_ofses[y*2] = y2;
			y3 /= dst_rect.height*4;
			bi->src_y_ofses[y*2+1] = y3;
			int gamma = GAMMA_TABLE_COUNT;
			int y2diff = abs((y2*2+1)*dst_rect.height*2-y1),
				y3diff = abs((y3*2+1)*dst_rect.height*2-y1);
			if (y2diff+y3diff > 0) {
				gamma = (GAMMA_TABLE_COUNT*y3diff)/(y2diff+y3diff);
			}
			bi->gamma_y[y] = (gamma*GAMMA_TABLE_VALUE)/GAMMA_TABLE_COUNT;
		}
	}
	bi->src_x_ofses.resize(draw_cx*2, 0);
	bi->gamma_x.resize(draw_cx, 0);
	if (dst_rect.width >= src_rect.width) {
		int base_x = src_rect.x*dst_rect.width*2,
			max_x = src_rect.width*dst_rect.width*2;
		for (int x = 0; x < draw_cx; x++) {
			int x1 = ((x+draw_x)*2+1)*src_rect.width+base_x,
				x2 = x1-dst_rect.width;
			if (x2 < 0) {
				x2 = 0;
			} else if (x2 >= max_x) {
				x2 = max_x-1;
			}
			int gamma = GAMMA_TABLE_COUNT-
				((x2%(dst_rect.width*2))*GAMMA_TABLE_COUNT)/
					(dst_rect.width*2);
			chaos_uint32 gamma_xa = gamma_table[gamma];
			bi->gamma_x[x] = gamma_xa |
				((GAMMA_TABLE_VALUE-gamma_xa) << 16);
			x2 /= dst_rect.width*2;
			bi->src_x_ofses[x*2] = x2;
			if (x2 < src_rect.width-1) {
				x2++;
			}
			bi->src_x_ofses[x*2+1] = x2;
		}
	} else {
		int base_x = src_rect.x*dst_rect.width*4,
			max_x = src_rect.width*dst_rect.width*4;
		for (int x = 0; x < draw_cx; x++) {
			int x1 = ((x+draw_x)*2+1)*src_rect.width*2+base_x,
				x2 = x1-src_rect.width,
				x3 = x1+src_rect.width;
			if (x2 < 0) {
				x2 = 0;
			} else if (x2 >= max_x) {
				x2 = max_x-1;
			}
			if (x3 < 0) {
				x3 = 0;
			} else if (x3 >= max_x) {
				x3 = max_x-1;
			}
			x2 /= dst_rect.width*4;
			bi->src_x_ofses[x*2] = x2;
			x3 /= dst_rect.width*4;
			bi->src_x_ofses[x*2+1] = x3;
			int gamma = GAMMA_TABLE_COUNT;
			int x2diff = abs((x2*2+1)*dst_rect.width*2-x1),
				x3diff = abs((x3*2+1)*dst_rect.width*2-x1);
			if (x2diff+x3diff > 0) {
				gamma = (GAMMA_TABLE_COUNT*x3diff)/(x2diff+x3diff);
			}
			chaos_uint32 gamma_xa =
				(gamma*GAMMA_TABLE_VALUE)/GAMMA_TABLE_COUNT;
			bi->gamma_x[x] = gamma_xa |
				((GAMMA_TABLE_VALUE-gamma_xa) << 16);
		}
	}
}

// fast smooth stretch blt(zoom in/out ok)

void chaos_image::fast_smooth_stretch_blt(
	fast_smooth_stretch_blt_info* bi,
	const chaos_image* blt_image,
	chaos_pos blt_pos,
	chaos_dim blt_dim,
	const chaos_rect* clip_rect)
{
	std::auto_ptr<fast_smooth_stretch_blt_info> bi_holder;
	if (bi == NULL) {
		bi_holder = std::auto_ptr<fast_smooth_stretch_blt_info>(
			new fast_smooth_stretch_blt_info());
		bi = bi_holder.get();
	}
	chaos_rect clip_rect2(chaos_pos(0, 0), get_dim());
	if (clip_rect != NULL) {
		clip_rect2 &= *clip_rect;
	}
	setup_fast_smooth_stretch_blt_info(
		bi,
		chaos_rect(blt_pos, blt_dim),
		chaos_rect(chaos_pos(0, 0), blt_image->get_dim()),
		clip_rect2);
	if ((bi->blt_dst_rect.width <= 0) ||
		(bi->blt_dst_rect.height <= 0))
	{
		return;
	}
	chaos_uint32* dst_ptr = get_bits(bi->blt_dst_rect.get_lt_pos());
	chaos_uint32* src_y_ofses = &bi->src_y_ofses[0];
	chaos_uint32* src_x_ofses = &bi->src_x_ofses[0];
	chaos_uint32* gamma_y = &bi->gamma_y[0];
	chaos_uint32* gamma_x = &bi->gamma_x[0];
	int draw_cx = bi->blt_dst_rect.width,
		draw_cy = bi->blt_dst_rect.height,
		x, y;
	if (bi->same_size) {
		const chaos_uint32* src_ptr_a =
			blt_image->get_bits(src_x_ofses[0], src_y_ofses[0]);
		int length = draw_cx*4;
		if ((draw_cx == get_scanline_length()) &&
			(draw_cx == blt_image->get_scanline_length()))
		{
			memcpy(dst_ptr, src_ptr_a, length*draw_cy);
		} else {
			for (y = 0; y < draw_cy; y++) {
				memcpy(dst_ptr, src_ptr_a, length);
				dst_ptr += get_scanline_length();
				src_ptr_a += blt_image->get_scanline_length();
			}
		}
	} else {
		for (y = 0; y < draw_cy; y++) {
			const chaos_uint32* src_ptr_a =
				blt_image->get_bits(0, src_y_ofses[y*2]);
			const chaos_uint32* src_ptr_b =
				blt_image->get_bits(0, src_y_ofses[y*2+1]);
			chaos_uint32 gamma_ya = gamma_y[y],
				gamma_yb = GAMMA_TABLE_VALUE-gamma_ya;
			for (x = 0; x < draw_cx; x++) {
				chaos_uint32 src_x_ofs_a = src_x_ofses[x*2],
					src_y_ofs_b = src_x_ofses[x*2+1],
					gamma_xab = gamma_x[x],
					gamma_ab = gamma_xab*gamma_ya,
					gamma_cd = gamma_xab*gamma_yb,
					delta_a = gamma_ab & 0xFFFF,
					delta_b = gamma_ab >> 16,
					delta_c = gamma_cd & 0xFFFF,
					delta_d = gamma_cd >> 16,
					pixel_a = src_ptr_a[src_x_ofs_a],
					pixel_b = src_ptr_a[src_y_ofs_b],
					pixel_c = src_ptr_b[src_x_ofs_a],
					pixel_d = src_ptr_b[src_y_ofs_b];
				dst_ptr[x] = (chaos_uint32)(
					(((((pixel_a & 0x00FF00FF)*delta_a)+
						((pixel_b & 0x00FF00FF)*delta_b)+
						((pixel_c & 0x00FF00FF)*delta_c)+
						((pixel_d & 0x00FF00FF)*delta_d))
							>> 8) & 0x00FF00FF) |
					(((((pixel_a & 0xFF00FF00) >> 8)*delta_a)+
						(((pixel_b & 0xFF00FF00) >> 8)*delta_b)+
						(((pixel_c & 0xFF00FF00) >> 8)*delta_c)+
						(((pixel_d & 0xFF00FF00) >> 8)*delta_d))
							& 0xFF00FF00));
			}
			dst_ptr += get_scanline_length();
		}
	}
}

// blt width alpha

void chaos_image::alpha_blt(
	const chaos_image* blt_image,
	const chaos_image* alpha_image,
	chaos_pos blt_pos,
	const chaos_rect* clip_rect)
{
	chaos_rect blt_rect(0, 0, width, height);
	blt_rect &= chaos_rect(blt_pos, blt_image->get_dim());
	blt_rect &= chaos_rect(blt_pos, alpha_image->get_dim());
	if (clip_rect != NULL) {
		blt_rect &= *clip_rect;
	}
	if (!blt_rect.empty()) {
		chaos_uint32* dst_bits =
			get_bits(blt_rect.get_lt_pos());
		const chaos_uint32* src_bits =
			blt_image->get_bits(blt_rect.get_lt_pos()-blt_pos);
		const chaos_uint8* alpha_bits = (chaos_uint8*)
			alpha_image->get_bits(blt_rect.get_lt_pos()-blt_pos)+
				CHAOS_RGBA_ALPHA_OFS;
		int dst_step = scanline_length-blt_rect.width,
			src_step = blt_image->get_scanline_length()-blt_rect.width,
			alpha_step =
				(alpha_image->get_scanline_length()-blt_rect.width)*4;
		int y = blt_rect.height;
		do {
			int x = blt_rect.width;
			do {
				chaos_uint32 dst_col = *dst_bits,
					src_col = *(src_bits++);
				int alpha = *alpha_bits;
				alpha_bits += 4;
				alpha += (alpha+1)/256;
				int alpha2 = 256-alpha;
				*(dst_bits++) = CHAOS_RGBA(
					(CHAOS_RGBA_RED(dst_col)*alpha+
						CHAOS_RGBA_RED(src_col)*alpha2)/256,
					(CHAOS_RGBA_GREEN(dst_col)*alpha+
						CHAOS_RGBA_GREEN(src_col)*alpha2)/256,
					(CHAOS_RGBA_BLUE(dst_col)*alpha+
						CHAOS_RGBA_BLUE(src_col)*alpha2)/256,
					0);
			} while (--x != 0);
			dst_bits += dst_step;
			src_bits += src_step;
			alpha_bits += alpha_step;
		} while (--y != 0);
	}
}

// blt width alpha(single image)

void chaos_image::alpha_blt(
	const chaos_image* blt_image,
	chaos_pos blt_pos,
	const chaos_rect* clip_rect)
{
	chaos_rect blt_rect(0, 0, width, height);
	blt_rect &= chaos_rect(blt_pos, blt_image->get_dim());
	if (clip_rect != NULL) {
		blt_rect &= *clip_rect;
	}
	if (!blt_rect.empty()) {
		chaos_uint32* dst_bits =
			get_bits(blt_rect.get_lt_pos());
		const chaos_uint32* src_bits =
			blt_image->get_bits(blt_rect.get_lt_pos()-blt_pos);
		int dst_step = scanline_length-blt_rect.width,
			src_step = blt_image->get_scanline_length()-blt_rect.width;
		int y = blt_rect.height;
		do {
			int x = blt_rect.width;
			do {
				chaos_uint32 dst_col = *dst_bits,
					src_col = *(src_bits++);
				int alpha = CHAOS_RGBA_ALPHA(src_col);
				alpha += (alpha+1)/256;
				int alpha2 = 256-alpha;
				*(dst_bits++) = CHAOS_RGBA(
					(CHAOS_RGBA_RED(dst_col)*alpha+
						CHAOS_RGBA_RED(src_col)*alpha2)/256,
					(CHAOS_RGBA_GREEN(dst_col)*alpha+
						CHAOS_RGBA_GREEN(src_col)*alpha2)/256,
					(CHAOS_RGBA_BLUE(dst_col)*alpha+
						CHAOS_RGBA_BLUE(src_col)*alpha2)/256,
					0);
			} while (--x != 0);
			dst_bits += dst_step;
			src_bits += src_step;
		} while (--y != 0);
	}
}

// blt width alpha

void chaos_image::mask_blt(
	const chaos_image* blt_image,
	const chaos_image* mask_image,
	chaos_pos blt_pos,
	const chaos_rect* clip_rect)
{
	chaos_rect blt_rect(0, 0, width, height);
	blt_rect &= chaos_rect(blt_pos, blt_image->get_dim());
	blt_rect &= chaos_rect(blt_pos, mask_image->get_dim());
	if (clip_rect != NULL) {
		blt_rect &= *clip_rect;
	}
	if (!blt_rect.empty()) {
		chaos_uint32* dst_bits =
			get_bits(blt_rect.get_lt_pos());
		const chaos_uint32* src_bits =
			blt_image->get_bits(blt_rect.get_lt_pos()-blt_pos);
		const chaos_uint8* mask_bits = (chaos_uint8*)
			mask_image->get_bits(blt_rect.get_lt_pos()-blt_pos)+
				CHAOS_RGBA_ALPHA_OFS;
		int dst_step = scanline_length-blt_rect.width,
			src_step = blt_image->get_scanline_length()-blt_rect.width,
			mask_step =
				(mask_image->get_scanline_length()-blt_rect.width)*4;
		int y = blt_rect.height;
		do {
			int x = blt_rect.width;
			do {
				chaos_uint32 dst_col = *dst_bits,
					src_col = *(src_bits++);
				int alpha = *mask_bits;
				mask_bits += 4;
				int alpha2 = 255-alpha;
				*(dst_bits++) = CHAOS_RGBA(
					(CHAOS_RGBA_RED(dst_col) & alpha) |
						(CHAOS_RGBA_RED(src_col) & alpha2),
					(CHAOS_RGBA_GREEN(dst_col) & alpha) |
						(CHAOS_RGBA_GREEN(src_col)& alpha2),
					(CHAOS_RGBA_BLUE(dst_col) & alpha) |
						(CHAOS_RGBA_BLUE(src_col) & alpha2),
					0);
			} while (--x != 0);
			dst_bits += dst_step;
			src_bits += src_step;
			mask_bits += mask_step;
		} while (--y != 0);
	}
}

// blt width mask(single image)

void chaos_image::mask_blt(
	const chaos_image* blt_image,
	chaos_pos blt_pos,
	const chaos_rect* clip_rect)
{
	chaos_rect blt_rect(0, 0, width, height);
	blt_rect &= chaos_rect(blt_pos, blt_image->get_dim());
	if (clip_rect != NULL) {
		blt_rect &= *clip_rect;
	}
	if (!blt_rect.empty()) {
		chaos_uint32* dst_bits =
			get_bits(blt_rect.get_lt_pos());
		const chaos_uint32* src_bits =
			blt_image->get_bits(blt_rect.get_lt_pos()-blt_pos);
		int dst_step = scanline_length-blt_rect.width,
			src_step = blt_image->get_scanline_length()-blt_rect.width;
		int y = blt_rect.height;
		do {
			int x = blt_rect.width;
			do {
				chaos_uint32 dst_col = *dst_bits,
					src_col = *(src_bits++);
				int alpha = CHAOS_RGBA_ALPHA(src_col);
				int alpha2 = 255-alpha;
				*(dst_bits++) = CHAOS_RGBA(
					(CHAOS_RGBA_RED(dst_col) & alpha) |
						(CHAOS_RGBA_RED(src_col) & alpha2),
					(CHAOS_RGBA_GREEN(dst_col) & alpha) |
						(CHAOS_RGBA_GREEN(src_col)& alpha2),
					(CHAOS_RGBA_BLUE(dst_col) & alpha) |
						(CHAOS_RGBA_BLUE(src_col) & alpha2),
					0);
			} while (--x != 0);
			dst_bits += dst_step;
			src_bits += src_step;
		} while (--y != 0);
	}
}

// fade

void chaos_image::fade_blt(
	const chaos_image* blt_image,
	int fade,
	chaos_pos blt_pos,
	const chaos_rect* clip_rect)
{
	if (fade >= 255) {
		return;
	}
	if (fade <= 0) {
		blt(blt_image, blt_pos, clip_rect);
		return;
	}
	int alpha = fade+(fade+1)/256;
	int alpha2 = 256-alpha;
	chaos_rect blt_rect(0, 0, width, height);
	blt_rect &= chaos_rect(blt_pos, blt_image->get_dim());
	if (clip_rect != NULL) {
		blt_rect &= *clip_rect;
	}
	if (!blt_rect.empty()) {
		chaos_uint32* dst_bits =
			get_bits(blt_rect.get_lt_pos());
		const chaos_uint32* src_bits =
			blt_image->get_bits(blt_rect.get_lt_pos()-blt_pos);
		int dst_step = scanline_length-blt_rect.width,
			src_step = blt_image->get_scanline_length()-blt_rect.width;
		int y = blt_rect.height;
		do {
			int x = blt_rect.width;
			do {
				chaos_uint32 dst_col = *dst_bits,
					src_col = *(src_bits++);
				*(dst_bits++) = CHAOS_RGBA(
					(CHAOS_RGBA_RED(dst_col)*alpha+
						CHAOS_RGBA_RED(src_col)*alpha2)/256,
					(CHAOS_RGBA_GREEN(dst_col)*alpha+
						CHAOS_RGBA_GREEN(src_col)*alpha2)/256,
					(CHAOS_RGBA_BLUE(dst_col)*alpha+
						CHAOS_RGBA_BLUE(src_col)*alpha2)/256,
					0);
			} while (--x != 0);
			dst_bits += dst_step;
			src_bits += src_step;
		} while (--y != 0);
	}
}

// horizontal blind blt

void chaos_image::horz_blind_blt(
	const chaos_image* blind_image,
	int gap, int fat,
	const chaos_rect* clip_rect)
{
	if ((gap == 0) || (fat <= 0)) {
		return;
	}
	int agap = abs(gap);
	if (fat >= agap) {
		blt(blind_image, chaos_pos(0, 0), clip_rect);
		return;
	}
	chaos_rect blt_rect(0, 0, width, height);
	blt_rect &= chaos_rect(chaos_pos(0, 0), blind_image->get_dim());
	if (clip_rect != NULL) {
		blt_rect &= *clip_rect;
	}
	if (!blt_rect.empty()) {
		chaos_uint32* dst_bits =
			get_bits(blt_rect.get_lt_pos());
		const chaos_uint32* src_bits =
			blind_image->get_bits(blt_rect.get_lt_pos());
		int dst_step = scanline_length-blt_rect.width,
			src_step = blind_image->get_scanline_length()-blt_rect.width;
		int y_start = blt_rect.y,
			y_end = blt_rect.y+blt_rect.height;
		if (gap > 0) {
			int y = y_start;
			do {
				if (y%agap < fat) {
					int x = blt_rect.width;
					do {
						*(dst_bits++) = *(src_bits++);
					} while (--x > 0);
				} else {
					dst_bits += blt_rect.width;
					src_bits += blt_rect.width;
				}
				dst_bits += dst_step;
				src_bits += src_step;
			} while (++y < y_end);
		} else {
			int fat2 = agap-fat;
			int y = y_start;
			do {
				if (y%agap >= fat2) {
					int x = blt_rect.width;
					do {
						*(dst_bits++) = *(src_bits++);
					} while (--x > 0);
				} else {
					dst_bits += blt_rect.width;
					src_bits += blt_rect.width;
				}
				dst_bits += dst_step;
				src_bits += src_step;
			} while (++y < y_end);
		}
	}
}

// vertical blind blt

void chaos_image::vert_blind_blt(
	const chaos_image* blind_image,
	int gap, int fat,
	const chaos_rect* clip_rect)
{
	if ((gap == 0) || (fat <= 0)) {
		return;
	}
	int agap = abs(gap);
	if (fat >= agap) {
		blt(blind_image, chaos_pos(0, 0), clip_rect);
		return;
	}
	chaos_rect blt_rect(0, 0, width, height);
	blt_rect &= chaos_rect(chaos_pos(0, 0), blind_image->get_dim());
	if (clip_rect != NULL) {
		blt_rect &= *clip_rect;
	}
	if (!blt_rect.empty()) {
		chaos_uint32* dst_bits =
			get_bits(blt_rect.get_lt_pos());
		const chaos_uint32* src_bits =
			blind_image->get_bits(blt_rect.get_lt_pos());
		int dst_step = scanline_length-blt_rect.width,
			src_step = blind_image->get_scanline_length()-blt_rect.width;
		int x_start = blt_rect.x,
			x_end = blt_rect.x+blt_rect.width;
		if (gap > 0) {
			int y = blt_rect.height;
			do {
				int x = x_start;
				do {
					if (x%agap < fat) {
						*dst_bits = *src_bits;
					}
					dst_bits++;
					src_bits++;
				} while (++x < x_end);
				dst_bits += dst_step;
				src_bits += src_step;
			} while (--y > 0);
		} else {
			int fat2 = agap-fat;
			int y = blt_rect.height;
			do {
				int x = x_start;
				do {
					if (x%agap >= fat2) {
						*dst_bits = *src_bits;
					}
					dst_bits++;
					src_bits++;
				} while (++x < x_end);
				dst_bits += dst_step;
				src_bits += src_step;
			} while (--y > 0);
		}
	}
}
